#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Placeholder for adding c (or ctypes) extensions to PsychoPy on linux
"""
from __future__ import absolute_import, print_function


def rush(value=True):
    """Raise the priority of the current thread/process.
    NOT YET IMPLEMENTED ON posix (but you can use
    > nice python ......

    """
    return False
