# -*- coding: utf-8 -*-
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
"""ioHub Common Eye Tracker Interface for Tobii (C) Eye Tracking System"""

from __future__ import absolute_import

from .eyetracker import *
