gitIgnoreText = """
# Byte-compiled / optimized / DLL files
__pycache__/
*.pyc
*.pyo
*.pyd
*.so

# Backup files
*.bak
~$*.xls*
~$*.doc*
~$*.ppt*

# Jupyter Notebook
.ipynb_checkpoints

# Virtual Environment files
.env
.venv
env/
venv/
ENV/

# Spyder project settings
.spyderproject
.spyproject

# OS generated files
.DS_Store
.directory
.gdb_history
ehthumbs.db
Icon?
*.orig
old
Thumbs.db
.Spotlight-V100
.Trashes

# lib files used for local debugging
/lib/
/html/lib/

"""
